==> montecarlodata-0.70.0-py3-none-any.whl: montecarlodata/agents/agent.py <==
import json
import re
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import (
    List,
    Optional,
)
from urllib.parse import urljoin

import click
import questionary
import requests
from dataclasses_json import (
    DataClassJsonMixin,
    Undefined,
    dataclass_json,
)
from pycarlo.core import (
    Client,
    Mutation,
    Query,
)
from tabulate import tabulate

from montecarlodata import settings
from montecarlodata.agents.fields import (
    AWS,
    AZURE_STORAGE_ACCOUNT_KEYS,
    GCP_JSON_SERVICE_ACCOUNT_KEY,
    AWS_ASSUMABLE_ROLE,
    REMOTE_AGENT,
)
from montecarlodata.common.common import read_as_json_string
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors, complain_and_abort
from montecarlodata.utils import AwsClientWrapper

DEFAULT_TAGS_PAGE_SIZE = 10
DOCKER_HUB_TAGS_BASE_URL = "https://hub.docker.com/v2/namespaces/"

# Adapted from https://docs.aws.amazon.com/lambda/latest/dg/API_Invoke.html#API_Invoke_RequestSyntax
LAMBDA_ARN_REGEX = re.compile(
    r"^arn:(aws[a-zA-Z-]*)?:lambda:([a-z]{2}(-gov)?-[a-z]+-\d{1}):\d{12}:function:([a-zA-Z0-9-_\.]+)(:(\$LATEST|[a-zA-Z0-9-_]+))?$"
)


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass()
class DockerTag(DataClassJsonMixin):
    name: str
    last_updated: str


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass()
class DockerTagsResult(DataClassJsonMixin):
    results: List[DockerTag]


class AgentService:
    _AGENT_FRIENDLY_HEADERS = [
        "Agent ID",
        "Agent Type / Platform",
        "DC ID",
        "Endpoint",
        "Version",
        "Last updated (UTC)",
        "Active",
    ]

    def __init__(
        self,
        config: Config,
        mc_client: Client,
        aws_wrapper: Optional[AwsClientWrapper] = None,
        user_service: Optional[UserService] = None,
    ):
        self._abort_on_error = True
        self._mc_client = mc_client
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(
            profile_name=config.aws_profile,
            region_name=config.aws_region,
        )
        self._user_service = user_service or UserService(config=config)
        self._image_org = config.mcd_agent_image_org
        self._image_repo = config.mcd_agent_image_repo
        self._image_name = (
            f"{config.mcd_agent_image_host}/{self._image_org}/{self._image_repo}"
        )
        self._ecr_repo = config.mcd_agent_ecr_repo
        self._ecr_registry_id = config.mcd_agent_ecr_registry_id
        self._ecr_image_name = (
            f"{self._ecr_registry_id}.dkr.ecr.*.amazonaws.com/{self._ecr_repo}"
        )

    @manage_errors
    def create_agent(
        self, agent_type, platform, storage, auth_type, endpoint, **kwargs
    ) -> None:
        """
        Register an agent by validating connection and creating an AgentModel in the monolith.
        """

        dry_run = kwargs.get("dry_run", False)
        agent_request = {
            "agent_type": agent_type,
            "endpoint": endpoint,
            "storage_type": storage,
            "platform": platform,
            "auth_type": auth_type,
            "dry_run": dry_run,
        }

        if kwargs.get("dc_id"):
            agent_request["data_collector_id"] = kwargs["dc_id"]

        aws_region: Optional[str] = kwargs.get("aws_region")
        if platform == AWS and agent_type == REMOTE_AGENT and not aws_region:
            matches = LAMBDA_ARN_REGEX.match(endpoint)
            if matches:
                aws_region = matches.group(2)
        if auth_type == GCP_JSON_SERVICE_ACCOUNT_KEY:
            agent_request["credentials"] = read_as_json_string(kwargs["key_file"])
        elif auth_type == AWS_ASSUMABLE_ROLE:
            creds = {"aws_assumable_role": kwargs["assumable_role"]}
            if kwargs["external_id"]:
                creds["external_id"] = kwargs["external_id"]
            if aws_region:
                creds["aws_region"] = aws_region
            agent_request["credentials"] = json.dumps(creds)
        elif auth_type == AZURE_STORAGE_ACCOUNT_KEYS:
            creds = {"azure_connection_string": kwargs["connection_string"]}
            agent_request["credentials"] = json.dumps(creds)

        mutation = Mutation()
        # the trailing call to __fields__ is needed to force selection of all possible fields
        mutation.create_or_update_agent(**agent_request).__fields__()
        result = self._mc_client(mutation).create_or_update_agent

        self._validate_response(result.validation_result)

        if result.agent_id is not None:
            click.echo("Agent successfully registered!\n" f"AgentId: {result.agent_id}")
        elif dry_run:
            if result.validation_result.success:
                click.echo("Dry run completed successfully!")
            else:
                complain_and_abort("Dry run failed.")
        else:
            complain_and_abort("Failed to register agent.")

    @manage_errors
    def delete_agent(self, agent_id) -> None:
        """
        Deregister an Agent (deletes AgentModel from monolith)
        """
        variables = dict(agent_id=agent_id)

        mutation = Mutation()
        mutation.delete_agent(**variables)
        result = self._mc_client(mutation).delete_agent

        if result.success:
            click.echo(f"Agent {agent_id} deregistered.")
        else:
            complain_and_abort("Failed to deregister agent.")

    @manage_errors
    def echo_agents(
        self,
        show_inactive: bool = False,
        headers: Optional[str] = "firstrow",
        table_format: Optional[str] = "fancy_grid",
    ):
        """
        Display agents in an easy-to-read table.
        """

        table = [self._AGENT_FRIENDLY_HEADERS]
        for agent in self._user_service.agents:
            is_active = not agent.get("isDeleted")
            if not show_inactive and not is_active:
                continue
            full_type = f"{agent.get('agentType', '')} / {agent.get('platform', '')}"
            last_updated = (
                datetime.fromisoformat(agent.get("lastUpdatedTime")).strftime(
                    "%Y-%m-%d %H:%M:%S"
                )
                if agent.get("lastUpdatedTime")
                else "-"
            )

            table += [
                [
                    agent.get("uuid") or "",
                    full_type,
                    agent.get("dc_id") or "",
                    agent.get("endpoint") or "",
                    agent.get("imageVersion") or "-",
                    last_updated,
                    is_active,
                ]
            ]

        # If the account has no agents, add 1 line of empty values so tabulate() creates a pretty
        # empty table
        if len(table) == 1:
            table += ["" for _ in self._AGENT_FRIENDLY_HEADERS]

        click.echo(
            tabulate(table, headers=headers, tablefmt=table_format, maxcolwidths=100)
        )

    @staticmethod
    def _validate_response(validation_result):
        output = {}

        if validation_result.errors:
            errors = []
            for error in validation_result.errors:
                error_output = dict(
                    message=error.friendly_message or "",
                    cause=error.cause or "",
                    resolution=error.resolution or "",
                )
                if settings.MCD_VERBOSE_ERRORS and error.stack_trace:
                    stack_trace_filename = f"mcd_error_trace_{uuid.uuid4()}.txt"
                    with open(stack_trace_filename, "w") as stack_trace_file:
                        stack_trace_file.write(f"-----------\n{error.stack_trace}\n")
                    error_output[
                        "stack_trace"
                    ] = f"Stack trace written to {stack_trace_filename}"

                errors.append(error_output)
            output["errors"] = errors
        if validation_result.warnings:
            output["warnings"] = [
                dict(
                    message=warning.friendly_message or "",
                    cause=warning.cause or "",
                    resolution=warning.resolution or "",
                )
                for warning in validation_result.warnings
            ]

        # If there are any errors or warnings returned, display them
        if "errors" in output or "warnings" in output:
            click.echo(json.dumps(output, indent=4))

    @manage_errors
    def check_agent_health(self, **kwargs):
        storage_query = Query()
        storage_query.test_data_store_reachability(**kwargs)
        storage_result = self._mc_client(
            storage_query,
            idempotent_request_id=str(uuid.uuid4()),
            timeout_in_seconds=40,  # let Monolith timeout first
        ).test_data_store_reachability

        if not storage_result.success:
            return self._validate_response(storage_result)
        agent = self._user_service.get_agent(**kwargs)
        if agent.get("agentType") != "REMOTE_AGENT":
            return click.echo("Agent health check succeeded!")

        agent_query = Query()
        agent_query.test_agent_reachability(**kwargs)
        agent_result = self._mc_client(
            agent_query,
            idempotent_request_id=str(uuid.uuid4()),
            timeout_in_seconds=40,  # let Monolith timeout first
        ).test_agent_reachability

        if agent_result.success:
            if agent_result.additional_data.returned_data:
                click.echo(
                    json.dumps(agent_result.additional_data.returned_data, indent=4)
                )
            return click.echo("Agent health check succeeded!")

        return self._validate_response(agent_result)

    @manage_errors
    def upgrade_agent(self, **kwargs):
        image = self._choose_image(**kwargs)
        click.echo(f"Upgrading agent with image '{image}'")
        variables = {
            "agent_id": kwargs["agent_id"],
            "image": image,
        }

        mutation = Mutation()
        mutation.upgrade_agent(**variables)
        result = self._mc_client(
            mutation,
            idempotent_request_id=str(uuid.uuid4()),
            timeout_in_seconds=40,  # let Monolith timeout first
        ).upgrade_agent

        click.echo("Upgrade succeeded!")
        if result.upgrade_result:
            click.echo(json.dumps(result.upgrade_result, indent=4))

    def _choose_image(
        self, agent_id: str, image_tag: Optional[str] = None
    ) -> Optional[str]:
        platform = self._user_service.get_agent(agent_id).get("platform")
        use_aws_ecr = platform == "AWS"

        if not image_tag or not self._validate_tag(image_tag, use_aws_ecr):
            versions = self._get_recent_versions(platform)
            image_tag = (
                versions[0]
                if not image_tag
                else questionary.select(
                    "Please choose a valid image tag", choices=versions
                ).ask()
            )

        if not image_tag:
            # user canceled prompt without choosing a response
            raise click.Abort()

        if use_aws_ecr:
            return f"{self._ecr_image_name}:{image_tag}"
        return f"{self._image_name}:{image_tag}"

    def _validate_tag(self, image_tag: str, use_aws_ecr: bool) -> bool:
        if use_aws_ecr:
            return self._aws_wrapper.ecr_image_exists(
                registry_id=self._ecr_registry_id,
                repository_name=self._ecr_repo,
                image_tag=image_tag,
            )
        image_path = f"{self._image_org}/repositories/{self._image_repo}"
        url: str = urljoin(DOCKER_HUB_TAGS_BASE_URL, f"{image_path}/tags/{image_tag}")
        response = requests.head(url=url)
        return response.status_code == 200

    def _get_recent_versions(self, platform: str) -> List[str]:
        if platform == "AWS":
            return self._aws_wrapper.list_ecr_image_tags(
                registry_id=self._ecr_registry_id,
                repository_name=self._ecr_repo,
                limit=DEFAULT_TAGS_PAGE_SIZE,
            )
        image_path = f"{self._image_org}/repositories/{self._image_repo}"
        url = urljoin(
            DOCKER_HUB_TAGS_BASE_URL,
            f"{image_path}/tags?page_size={DEFAULT_TAGS_PAGE_SIZE}&page=1",
        )
        response = requests.get(url=url)
        response.raise_for_status()
        payload: DockerTagsResult = DockerTagsResult.from_json(response.text)
        image_variant = "cloudrun" if platform == "GCP" else "generic"
        click.echo(
            f"Defaulting to {image_variant} image variants because platform is {platform}"
        )
        payload.results = list(
            filter(
                lambda t: image_variant in t.name and "latest" not in t.name,
                payload.results,
            )
        )
        payload.results.sort(key=lambda t: t.last_updated, reverse=True)
        return [tag.name for tag in payload.results]

==> montecarlodata-0.70.2-py3-none-any.whl: montecarlodata/agents/agent.py <==
import json
import re
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import (
    List,
    Optional,
)
from urllib.parse import urljoin

import click
import questionary
import requests
from dataclasses_json import (
    DataClassJsonMixin,
    Undefined,
    dataclass_json,
)
from pycarlo.core import (
    Client,
    Mutation,
    Query,
)
from tabulate import tabulate

from montecarlodata import settings
from montecarlodata.agents.fields import (
    AWS,
    AZURE_STORAGE_ACCOUNT_KEYS,
    GCP_JSON_SERVICE_ACCOUNT_KEY,
    AWS_ASSUMABLE_ROLE,
    REMOTE_AGENT,
)
from montecarlodata.common.common import read_as_json_string
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors, complain_and_abort
from montecarlodata.utils import AwsClientWrapper

DEFAULT_TAGS_PAGE_SIZE = 10
DOCKER_HUB_TAGS_BASE_URL = "https://hub.docker.com/v2/namespaces/"

# Adapted from https://docs.aws.amazon.com/lambda/latest/dg/API_Invoke.html#API_Invoke_RequestSyntax
LAMBDA_ARN_REGEX = re.compile(
    r"^arn:(aws[a-zA-Z-]*)?:lambda:([a-z]{2}(-gov)?-[a-z]+-\d{1}):\d{12}:function:([a-zA-Z0-9-_\.]+)(:(\$LATEST|[a-zA-Z0-9-_]+))?$"
)


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass()
class DockerTag(DataClassJsonMixin):
    name: str
    last_updated: str


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass()
class DockerTagsResult(DataClassJsonMixin):
    results: List[DockerTag]


class AgentService:
    _AGENT_FRIENDLY_HEADERS = [
        "Agent ID",
        "Agent Type / Platform",
        "DC ID",
        "Endpoint",
        "Version",
        "Last updated (UTC)",
        "Active",
    ]

    def __init__(
        self,
        config: Config,
        mc_client: Client,
        aws_wrapper: Optional[AwsClientWrapper] = None,
        user_service: Optional[UserService] = None,
    ):
        self._abort_on_error = True
        self._mc_client = mc_client
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(
            profile_name=config.aws_profile,
            region_name=config.aws_region,
        )
        self._user_service = user_service or UserService(config=config)
        self._image_org = config.mcd_agent_image_org
        self._image_repo = config.mcd_agent_image_repo
        self._image_name = (
            f"{config.mcd_agent_image_host}/{self._image_org}/{self._image_repo}"
        )
        self._ecr_repo = config.mcd_agent_ecr_repo
        self._ecr_registry_id = config.mcd_agent_ecr_registry_id
        self._ecr_image_name = (
            f"{self._ecr_registry_id}.dkr.ecr.*.amazonaws.com/{self._ecr_repo}"
        )

    @manage_errors
    def create_agent(
        self, agent_type, platform, storage, auth_type, endpoint, **kwargs
    ) -> None:
        """
        Register an agent by validating connection and creating an AgentModel in the monolith.
        """

        dry_run = kwargs.get("dry_run", False)
        agent_request = {
            "agent_type": agent_type,
            "endpoint": endpoint,
            "storage_type": storage,
            "platform": platform,
            "auth_type": auth_type,
            "dry_run": dry_run,
        }

        if kwargs.get("dc_id"):
            agent_request["data_collector_id"] = kwargs["dc_id"]

        aws_region: Optional[str] = None
        if platform == AWS and agent_type == REMOTE_AGENT:
            matches = LAMBDA_ARN_REGEX.match(endpoint)
            if matches:
                aws_region = matches.group(2)
        if auth_type == GCP_JSON_SERVICE_ACCOUNT_KEY:
            agent_request["credentials"] = read_as_json_string(kwargs["key_file"])
        elif auth_type == AWS_ASSUMABLE_ROLE:
            creds = {"aws_assumable_role": kwargs["assumable_role"]}
            if kwargs["external_id"]:
                creds["external_id"] = kwargs["external_id"]
            if aws_region:
                creds["aws_region"] = aws_region
            agent_request["credentials"] = json.dumps(creds)
        elif auth_type == AZURE_STORAGE_ACCOUNT_KEYS:
            creds = {"azure_connection_string": kwargs["connection_string"]}
            agent_request["credentials"] = json.dumps(creds)

        mutation = Mutation()
        # the trailing call to __fields__ is needed to force selection of all possible fields
        mutation.create_or_update_agent(**agent_request).__fields__()
        result = self._mc_client(mutation).create_or_update_agent

        self._validate_response(result.validation_result)

        if result.agent_id is not None:
            click.echo("Agent successfully registered!\n" f"AgentId: {result.agent_id}")
        elif dry_run:
            if result.validation_result.success:
                click.echo("Dry run completed successfully!")
            else:
                complain_and_abort("Dry run failed.")
        else:
            complain_and_abort("Failed to register agent.")

    @manage_errors
    def delete_agent(self, agent_id) -> None:
        """
        Deregister an Agent (deletes AgentModel from monolith)
        """
        variables = dict(agent_id=agent_id)

        mutation = Mutation()
        mutation.delete_agent(**variables)
        result = self._mc_client(mutation).delete_agent

        if result.success:
            click.echo(f"Agent {agent_id} deregistered.")
        else:
            complain_and_abort("Failed to deregister agent.")

    @manage_errors
    def echo_agents(
        self,
        show_inactive: bool = False,
        headers: Optional[str] = "firstrow",
        table_format: Optional[str] = "fancy_grid",
    ):
        """
        Display agents in an easy-to-read table.
        """

        table = [self._AGENT_FRIENDLY_HEADERS]
        for agent in self._user_service.agents:
            is_active = not agent.get("isDeleted")
            if not show_inactive and not is_active:
                continue
            full_type = f"{agent.get('agentType', '')} / {agent.get('platform', '')}"
            last_updated = (
                datetime.fromisoformat(agent.get("lastUpdatedTime")).strftime(
                    "%Y-%m-%d %H:%M:%S"
                )
                if agent.get("lastUpdatedTime")
                else "-"
            )

            table += [
                [
                    agent.get("uuid") or "",
                    full_type,
                    agent.get("dc_id") or "",
                    agent.get("endpoint") or "",
                    agent.get("imageVersion") or "-",
                    last_updated,
                    is_active,
                ]
            ]

        # If the account has no agents, add 1 line of empty values so tabulate() creates a pretty
        # empty table
        if len(table) == 1:
            table += ["" for _ in self._AGENT_FRIENDLY_HEADERS]

        click.echo(
            tabulate(table, headers=headers, tablefmt=table_format, maxcolwidths=100)
        )

    @staticmethod
    def _validate_response(validation_result):
        output = {}

        if validation_result.errors:
            errors = []
            for error in validation_result.errors:
                error_output = dict(
                    message=error.friendly_message or "",
                    cause=error.cause or "",
                    resolution=error.resolution or "",
                )
                if settings.MCD_VERBOSE_ERRORS and error.stack_trace:
                    stack_trace_filename = f"mcd_error_trace_{uuid.uuid4()}.txt"
                    with open(stack_trace_filename, "w") as stack_trace_file:
                        stack_trace_file.write(f"-----------\n{error.stack_trace}\n")
                    error_output[
                        "stack_trace"
                    ] = f"Stack trace written to {stack_trace_filename}"

                errors.append(error_output)
            output["errors"] = errors
        if validation_result.warnings:
            output["warnings"] = [
                dict(
                    message=warning.friendly_message or "",
                    cause=warning.cause or "",
                    resolution=warning.resolution or "",
                )
                for warning in validation_result.warnings
            ]

        # If there are any errors or warnings returned, display them
        if "errors" in output or "warnings" in output:
            click.echo(json.dumps(output, indent=4))

    @manage_errors
    def check_agent_health(self, **kwargs):
        storage_query = Query()
        storage_query.test_data_store_reachability(**kwargs)
        storage_result = self._mc_client(
            storage_query,
            idempotent_request_id=str(uuid.uuid4()),
            timeout_in_seconds=40,  # let Monolith timeout first
        ).test_data_store_reachability

        if not storage_result.success:
            return self._validate_response(storage_result)
        agent = self._user_service.get_agent(**kwargs)
        if agent.get("agentType") != "REMOTE_AGENT":
            return click.echo("Agent health check succeeded!")

        agent_query = Query()
        agent_query.test_agent_reachability(**kwargs)
        agent_result = self._mc_client(
            agent_query,
            idempotent_request_id=str(uuid.uuid4()),
            timeout_in_seconds=40,  # let Monolith timeout first
        ).test_agent_reachability

        if agent_result.success:
            if agent_result.additional_data.returned_data:
                click.echo(
                    json.dumps(agent_result.additional_data.returned_data, indent=4)
                )
            return click.echo("Agent health check succeeded!")

        return self._validate_response(agent_result)

    @manage_errors
    def upgrade_agent(self, **kwargs):
        image = self._choose_image(**kwargs)
        click.echo(f"Upgrading agent with image '{image}'")
        variables = {
            "agent_id": kwargs["agent_id"],
            "image": image,
        }

        mutation = Mutation()
        mutation.upgrade_agent(**variables)
        result = self._mc_client(
            mutation,
            idempotent_request_id=str(uuid.uuid4()),
            timeout_in_seconds=40,  # let Monolith timeout first
        ).upgrade_agent

        click.echo("Upgrade succeeded!")
        if result.upgrade_result:
            click.echo(json.dumps(result.upgrade_result, indent=4))

    def _choose_image(
        self, agent_id: str, image_tag: Optional[str] = None
    ) -> Optional[str]:
        platform = self._user_service.get_agent(agent_id).get("platform")
        use_aws_ecr = platform == "AWS"

        if not image_tag or not self._validate_tag(image_tag, use_aws_ecr):
            versions = self._get_recent_versions(platform)
            image_tag = (
                versions[0]
                if not image_tag
                else questionary.select(
                    "Please choose a valid image tag", choices=versions
                ).ask()
            )

        if not image_tag:
            # user canceled prompt without choosing a response
            raise click.Abort()

        if use_aws_ecr:
            return f"{self._ecr_image_name}:{image_tag}"
        return f"{self._image_name}:{image_tag}"

    def _validate_tag(self, image_tag: str, use_aws_ecr: bool) -> bool:
        if use_aws_ecr:
            return self._aws_wrapper.ecr_image_exists(
                registry_id=self._ecr_registry_id,
                repository_name=self._ecr_repo,
                image_tag=image_tag,
            )
        image_path = f"{self._image_org}/repositories/{self._image_repo}"
        url: str = urljoin(DOCKER_HUB_TAGS_BASE_URL, f"{image_path}/tags/{image_tag}")
        response = requests.head(url=url)
        return response.status_code == 200

    def _get_recent_versions(self, platform: str) -> List[str]:
        if platform == "AWS":
            return self._aws_wrapper.list_ecr_image_tags(
                registry_id=self._ecr_registry_id,
                repository_name=self._ecr_repo,
                limit=DEFAULT_TAGS_PAGE_SIZE,
            )
        image_path = f"{self._image_org}/repositories/{self._image_repo}"
        url = urljoin(
            DOCKER_HUB_TAGS_BASE_URL,
            f"{image_path}/tags?page_size={DEFAULT_TAGS_PAGE_SIZE}&page=1",
        )
        response = requests.get(url=url)
        response.raise_for_status()
        payload: DockerTagsResult = DockerTagsResult.from_json(response.text)
        image_variant = "cloudrun" if platform == "GCP" else "generic"
        click.echo(
            f"Defaulting to {image_variant} image variants because platform is {platform}"
        )
        payload.results = list(
            filter(
                lambda t: image_variant in t.name and "latest" not in t.name,
                payload.results,
            )
        )
        payload.results.sort(key=lambda t: t.last_updated, reverse=True)
        return [tag.name for tag in payload.results]
